//#define _DEBUG
#include <math.h>
#include <propeller.h>
#include <sys/thread.h>
//**********************************//
#include "Motor_led_Driver.h"
#include "MPU6050_XS.h"
#ifdef _DEBUG
#include "FullDuplexSerial.h"
#endif
#include "sensfusion6.h"
#include "pid.h"
#include "controller.h"
#include "nRF24.h"
#include "RC.h"

//***********************************//
#define ATTITUDE_UPDATE_RATE_DIVIDER  2
#define FUSION_UPDATE_DT  (float)(1.0 / (IMU_UPDATE_FREQ / ATTITUDE_UPDATE_RATE_DIVIDER))
#define MOTORS_PWM_BITS     9
#define MOTORS_PWM_PERIOD   ((1<<MOTORS_PWM_BITS) - 1)
#define MOTORS_PWM_PRESCALE 0
#define C_BITS_TO_16(X) ((X)<<(16-MOTORS_PWM_BITS))
#define C_16_TO_BITS(X) ((X)>>(16-MOTORS_PWM_BITS)&((1<<MOTORS_PWM_BITS)-1))
#define STACK_SIZE 50
//***********************************//
static Axis3f gyro; // Gyro axis data in deg/s
static Axis3f acc;  // Accelerometer axis data 
static float eulerRollActual;
static float eulerPitchActual;
static float eulerYawActual;
static float eulerRollDesired;
static float eulerPitchDesired;
static float eulerYawDesired;
static float rollRateDesired;
static float pitchRateDesired;
static float yawRateDesired;
static float fusionDt;
static int cog3_stack[STACK_SIZE];

static _thread_state_t thread_data;
//**************************************//
typedef enum
{
  RATE,
  ANGLE
} RPYType;

RPYType rollType;
RPYType pitchType;
RPYType yawType;
//**************************************//
int loopTime;
int timer;
uint8_t  preflight_check[10];
uint8_t  Packettemplate[32];
uint8_t  attitudeCounter = 0;
int16_t  actuatorRoll;
int16_t  actuatorPitch;
int16_t  actuatorYaw;
uint16_t actuatorThrust;
uint16_t motorFrontLeft;
uint16_t motorFrontRight;
uint16_t motorMiddleLeft;
uint16_t motorMiddleRight;
uint16_t motorRearLeft;
uint16_t motorRearRight;
int16_t  f_ax;
int16_t  f_ay;
int16_t  f_az;
uint16_t rc_ch[9];
int8_t   rc_reciving;
//**************************************//
       void     pm_failsafe_loop(void);
       void     Pth_stabilizer_loop(void);
static void     distributePower(const uint16_t thrust, const int16_t roll, const int16_t pitch, const int16_t yaw);
static uint16_t limitThrust(int32_t value);
       uint8_t  Preeflight(void);
//**************************************//

int main(void)
{   
    waitcnt(CNT+CLKFREQ/5);
    // Start Full Duplex serial if later in Debug Mode
    #ifdef _DEBUG
    FullDuplexSerial_Start( 31,  30,  0,  115200);
    FullDuplexSerial_Str("H6 UAV Software V0.1\r");
    #endif
    if (Preeflight()==5){
        #ifdef _DEBUG
        FullDuplexSerial_Str("Preflight Check ok\r");
        #endif
        // From here start cog 3 with loop  for the stabilizer
       _start_cog_thread(cog3_stack + STACK_SIZE, Pth_stabilizer_loop, NULL, &thread_data);
    }
    else
    {
    //TODO: Errror Handling
    #ifdef _DEBUG 
    FullDuplexSerial_Str("Preflight Check Failed\r");
    #endif
    return;
    }
    // PM Loop Start should never exit 
    pm_failsafe_loop();   
    // Should never get here 
    return ; 
}

uint8_t Preeflight(void){

    // If normal Mode go start Preflight Check
    #ifdef _DEBUG
    FullDuplexSerial_Str("Starting Preflight check\r");
    #endif
    // Setup stuff we need to get of the ground
    rollRateDesired = 0;
    pitchRateDesired = 0;
    yawRateDesired = 0;
    actuatorThrust = 0;
    // Start Motor Driver in cog nr 2      
    if (Motor_led_Driver_Start()>0){
    preflight_check[0]=1;
    preflight_check[9]+=1;
    }
    // Start IMU/I2C Driver in cog nr 3
    if (MPU6050_XS_Start(17,16,21,20,0)>0){
    preflight_check[1]=1;
    preflight_check[9]+=1;
    }
    // Setup The controller PIDS
    if (controllerInit()>0){
    preflight_check[2]=1;
    preflight_check[9]+=1;
    }        
    if (nRF24_Start((int32_t)(&Packettemplate))>0){
    preflight_check[3]=1;
    preflight_check[9]+=1;
    }          
    // Setup RC Reciver PPM
    if (RC_Init(0, 1500 )>0){
    preflight_check[4]=1;
    preflight_check[9]+=1;
    }    
    return preflight_check[9];
}

void Pth_stabilizer_loop(void){

    while(1){
        timer = -CNT;
        // Start Time counter              
        MPU6050_XS_Imu6Read(&gyro, &acc);
        if (RC_Status()==-1){
        RC_Readall(&rc_ch);
        // Set Scale from Rc Ch 9
        float scale = 0.01 * (float)((rc_ch[8]-1500)/8);
        // Set Drift Comp from Rc Ch 8
        float drift = 0.01 * (float)((rc_ch[7]-1500)/4);
        eulerRollDesired = scale*(rc_ch[1]-1500);
        eulerPitchDesired =scale*(rc_ch[0]-1500);
        eulerYawDesired = (scale*(rc_ch[3]-1500))-drift;
        actuatorThrust = 65*(rc_ch[2]-1000);
        }
        else{
        eulerRollDesired = 0;
        eulerPitchDesired = 0;
        eulerYawDesired = 0;
        }
        
      if (++attitudeCounter >= ATTITUDE_UPDATE_RATE_DIVIDER)
      {
        sensfusion6UpdateQ(gyro.x, gyro.y, gyro.z, acc.x, acc.y, acc.z, FUSION_UPDATE_DT);

        sensfusion6GetEulerRPY(&eulerRollActual, &eulerPitchActual, &eulerYawActual);
        controllerCorrectAttitudePID(eulerRollActual, eulerPitchActual, eulerYawActual, eulerRollDesired, eulerPitchDesired, -eulerYawDesired, &rollRateDesired, &pitchRateDesired, &yawRateDesired);
        attitudeCounter = 0;
      }

        rollRateDesired = eulerRollDesired;

        pitchRateDesired = eulerPitchDesired;
 
        yawRateDesired = -eulerYawDesired;
   

      // TODO: Investigate possibility to subtract gyro drift.
      controllerCorrectRatePID(gyro.x, -gyro.y, gyro.z, rollRateDesired, pitchRateDesired, yawRateDesired);
      controllerGetActuatorOutput(&actuatorRoll, &actuatorPitch, &actuatorYaw);
      
      if (actuatorThrust > 1000)
      {   
      //Update 6 Motors       
      distributePower(actuatorThrust,actuatorPitch, actuatorRoll,  -actuatorYaw);          
      }else{
        distributePower(0, 0, 0, 0);
        controllerResetAllPID();
      } 
      // End of loop time mesure
      loopTime = (timer+CNT-16)/80000;
  
    }
}

static void distributePower(const uint16_t thrust, const int16_t roll, const int16_t pitch, const int16_t yaw)
{

  // Mix the Roll Pitch and Yaw to the 6 Motors in H config
  motorFrontLeft =      limitThrust(thrust + roll - pitch + yaw);
  motorFrontRight =     limitThrust(thrust - roll - pitch - yaw);
  motorMiddleLeft =     limitThrust(thrust + roll - yaw);
  motorMiddleRight =    limitThrust(thrust - roll + yaw);
  motorRearLeft =       limitThrust(thrust + roll + pitch + yaw);
  motorRearRight =      limitThrust(thrust - roll + pitch - yaw);
  // Set Motor Drivers      
  Motor_led_Driver_Pwm(13,C_16_TO_BITS(motorFrontLeft),511);
  Motor_led_Driver_Pwm(12,C_16_TO_BITS(motorFrontRight),511);
  Motor_led_Driver_Pwm(11,C_16_TO_BITS(motorMiddleLeft),511);
  Motor_led_Driver_Pwm(10,C_16_TO_BITS(motorMiddleRight),511);
  Motor_led_Driver_Pwm(9 ,C_16_TO_BITS(motorRearLeft),511);
  Motor_led_Driver_Pwm(8 ,C_16_TO_BITS(motorRearRight),511);
}

static uint16_t limitThrust(int32_t value)
{
  if(value > UINT16_MAX)
  {
    value = UINT16_MAX;
  }
  else if(value < 0)
  {
    value = 0;
  }

  return (uint16_t)value;
}

void pm_failsafe_loop(void){
    // Power Management loop cog 0 incl failsafe
    while(1){
    
        // Debug print out
        #ifdef _DEBUG
        waitcnt(CNT+CLKFREQ/20);
        FullDuplexSerial_Str("Loop time ms --");
        FullDuplexSerial_Dec(loopTime); 
        FullDuplexSerial_Str("\n"); 
        #endif
        
        waitcnt(CNT+CLKFREQ/1000);
        // Prepare telemetrie packet
        MPU6050_XS_Imu6(&f_ax, &f_ay, &f_az);
        // Start Sign for stream
        Packettemplate[0]= 0x7E ;
        // The Actua Euler angels for AFH
        Packettemplate[1]= (int16_t)eulerRollActual ;
        Packettemplate[2]= (int16_t)eulerRollActual>> 8 ;
        Packettemplate[3]= (int16_t)eulerPitchActual ;
        Packettemplate[4]= (int16_t)eulerPitchActual >> 8 ;
        Packettemplate[5]= (int16_t)eulerYawActual ;
        Packettemplate[6]= (int16_t)eulerYawActual >> 8 ;
        // Raw Accel Readings for raw AFH               
        Packettemplate[7]= (int16_t)f_ax ;
        Packettemplate[8]= (int16_t)f_ax >> 8 ;
        Packettemplate[9]= (int16_t)f_ay ;
        Packettemplate[10]= (int16_t)f_ay >> 8 ;
        Packettemplate[11]= (int16_t)f_az ;
        Packettemplate[12]= (int16_t)f_az >> 8 ;  
        // Motor Data ** mazbe better to send them later as 8 bit to show just the bar in ground station ?       
        Packettemplate[13]= (uint16_t)motorFrontLeft ;
        Packettemplate[14]= (uint16_t)motorFrontLeft >> 8 ;
        Packettemplate[15]= (uint16_t)motorMiddleLeft ;
        Packettemplate[16]= (uint16_t)motorMiddleLeft >> 8 ;
        Packettemplate[17]= (uint16_t)motorRearLeft ;
        Packettemplate[18]= (uint16_t)motorRearLeft >> 8 ;    
        Packettemplate[19]= (uint16_t)motorFrontRight ;
        Packettemplate[20]= (uint16_t)motorFrontRight >> 8 ;
        Packettemplate[21]= (uint16_t)motorMiddleRight ;
        Packettemplate[22]= (uint16_t)motorMiddleRight >> 8 ;
        Packettemplate[23]= (uint16_t)motorRearRight ;
        Packettemplate[24]= (uint16_t)motorRearRight >> 8 ;
        // Lipo Volatge
        Packettemplate[25]= 33 ;         
        
        // Power Management
        // 1 check ADC
        // 2 if below min start autodecent 
        
        // Fail Safe RC Reciver
        // If RC Reiver Failsafe fails, this Failsafe should kick in and Autoland
        // 1. If reciver malfunctions in midair set pitch roll yaw to 0
        // 2. Set Thrust to short below hover 
        // 3. If alttidue sensor check if we are on the ground and shut of enegines
         
    }
}